NOISE_LEVELS = [0, .1, .2, .3, .4, .5, .6, .7, .8, .9]

NIGEP_AVAILABLE_KWARGS = [
    'execution_name',
    'model',
    'batch_size',
    'input_shape',
    'x_data',
    'y_data',
    'target_names',
    'class_mode',
    'k_fold_n',
    'epochs',
    'callbacks',
    'noise_levels',
    'save_models',
    'evaluate_models',
]


from keras.models import Sequential


def train_model(model: Sequential, epochs, callbacks, train_data):
    x_train, y_train = train_data

    model.fit(
        x_train,
        y_train,
        verbose=0,
        callbacks=callbacks,
        epochs=epochs,
        validation_split=.1,
    )




from sklearn.metrics import classification_report, confusion_matrix
import numpy as np


def get_model_predictions(model, x_test, class_mode):
    predict_x = model.predict(x_test, verbose=0)
    if class_mode == 'binary':
        return np.round(predict_x).flatten().astype(int)

    return np.argmax(predict_x, axis=-1)


def get_confusion_matrix_and_report(y_test, predictions, target_names):
    y_true = np.argmax(y_test, axis=-1)
    cm = confusion_matrix(y_true, predictions)
    cr = classification_report(y_true, predictions, target_names=target_names, labels=np.arange(0,len(target_names),1))
    return cm, cr


def compute_metrics(model, class_mode, target_names, x_test, y_test):
    predictions = get_model_predictions(model, x_test, class_mode)
    cm, cr = get_confusion_matrix_and_report(y_test, predictions, target_names)
    return cm, cr



import os
import re
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from ..lib.functions import get_results_columns
from ..lib.folders import mkdir_output, get_directory_name

sns.set_theme(style="white")
sns.set_theme(style="whitegrid")
sns.set(font_scale=2)
sns.set_context("paper")
sns.set(font='serif')
sns.set_style("white", {
    "font.family": "serif",
    "font.serif": ["Times", "Palatino", "serif"]
})


class ResultsWriter:

    def __init__(self, name):
        mkdir_output()
        self.execution_folder_path = get_directory_name(f'{os.getcwd()}/output/{name}')
        os.mkdir(self.execution_folder_path)
        self.results_name = os.path.basename(self.execution_folder_path)

        self.heatmap_df = None
        self.mean_merged_df = None

    def __generate_df_by_csv(self, results_folder):
        df = pd.read_csv(results_folder + f'/results_{self.results_name}.csv')
        df.drop('Unnamed: 0', axis=1, inplace=True)
        return df

    def write_k_subset_folder(self, fold_number):
        results_folder = \
            f'{self.execution_folder_path}/kfold_{fold_number}__{datetime.now().isoformat().__str__()}'
        os.mkdir(results_folder)

        return results_folder

    def write_new_metrics(self, results_folder, train_noise, test_noise, cr, cm, target_names):
        pattern = re.compile(r'(\w+)\s+([\d.]+)\s+([\d.]+)\s+([\d.]+)\s+(\d+)')
        matches = pattern.findall(cr)

        columns = ['Class', 'Precision', 'Recall', 'F1-Score', 'Support']
        df = pd.DataFrame(matches, columns=columns).astype(
            {'Precision': float, 'Recall': float, 'F1-Score': float, 'Support': int})

        accuracy_match = re.search(r'accuracy\s+([\d.]+)', cr)
        accuracy = float(accuracy_match.group(1)) if accuracy_match else None

        classes_precision_dict = {}
        classes_recall_dict = {}
        classes_f1score_dict = {}
        for index, item in enumerate(target_names):
            classes_precision_dict[f'precision({item})'] = df['Precision'][index]
            classes_recall_dict[f'recall({item})'] = df['Recall'][index]
            classes_f1score_dict[f'f1-score({item})'] = df['F1-Score'][index]

        classes_number = len(target_names)
        metrics = {
            'train-dataset-noise': train_noise,
            'test-dataset-noise': test_noise,
            **classes_precision_dict,
            'precision(macro-avg)': df['Precision'][classes_number],
            'precision(weighted-avg)': df['Precision'][classes_number + 1],
            **classes_recall_dict,
            'recall(macro-avg)': df['Recall'][classes_number],
            'recall(weighted-avg)': df['Recall'][classes_number + 1],
            **classes_f1score_dict,
            'f1-score(accuracy)': accuracy,
            'f1-score(macro-avg)': df['F1-Score'][classes_number],
            'f1-score(weighted-avg)': df['F1-Score'][classes_number + 1],
        }

        if not os.path.isfile(results_folder + f'/results_{self.results_name}.csv'):
            (pd.DataFrame(metrics, index=[0], columns=get_results_columns(target_names))
             .to_csv(results_folder + f'/results_{self.results_name}.csv'))
            return

        current_df = self.__generate_df_by_csv(results_folder)

        pd.concat([current_df, pd.DataFrame(metrics, index=[0])]) \
            .to_csv(results_folder + f'/results_{self.results_name}.csv')

        # pd.DataFrame(cm).to_csv(
        #     results_folder + f'/train_{train_noise}_test_{test_noise}_confusion_matrix.csv'
        # )

    def save_mean_merged_results(self):
        unified_data = []
        train_noise = []
        test_noise = []
        for subdir, _, files in os.walk(f'{os.getcwd()}/output/{self.results_name}'):
            for file in files:
                if file.endswith('.csv') and file.startswith('results'):
                    file_path = os.path.join(subdir, file)
                    csv = pd.read_csv(file_path)
                    unified_data.append(csv['f1-score(weighted-avg)'])
                    train_noise = csv['train-dataset-noise']
                    test_noise = csv['test-dataset-noise']

        merged_df = pd.concat(unified_data, axis=1)

        mean_merged_data = {
            'train-noise': train_noise,
            'test-noise': test_noise,
            'f1-score(weighted-avg)': merged_df.mean(axis=1)
        }

        self.mean_merged_df = pd.DataFrame(mean_merged_data)

        self.mean_merged_df.to_csv(f'{os.getcwd()}/output/{self.results_name}/mean_results.csv')

    def save_heatmap_csv(self):
        column_values = self.mean_merged_df['train-noise'].unique()

        results_matrix = [
            self.mean_merged_df.loc[
                self.mean_merged_df['train-noise'] == value, 'f1-score(weighted-avg)'
            ].to_numpy() for value in column_values
        ]

        self.heatmap_df = pd.DataFrame(results_matrix, columns=column_values, index=column_values)
        self.heatmap_df.to_csv(f'{os.getcwd()}/output/{self.results_name}/generalization_profile_heatmap.csv')

    def plot_and_save_heatmap_png(self):
        plt.figure(figsize=(8, 6))
        sns.heatmap(self.heatmap_df,
                    cmap='coolwarm',
                    annot=True,
                    cbar_kws={'label': '5-Fold mean values of F-Score'},
                    fmt='.2f')

        plt.xlabel('Test Noise')
        plt.ylabel('Train Noise')

        plt.savefig(f'{os.getcwd()}/output/{self.results_name}/mean_results_heatmap.png')


from keras.models import Sequential
from sklearn.model_selection import KFold
from concurrent.futures import ThreadPoolExecutor
import threading

from .lib.apply_noise import apply_indexed_noise
from .lib.metrics import compute_metrics
from .lib.train_model import train_model
from .lib.consts import NOISE_LEVELS, NIGEP_AVAILABLE_KWARGS
from .classes.results_writer import ResultsWriter
from .lib.functions import validate_kwargs, write_model


class Nigep:

    def __init__(self, **kwargs):
        self.lock = threading.Lock()
        """
        Initialize Nigep instance.

        Parameters:
        - execution_name (str): Name of the execution.
        - model (Sequential): Keras Sequential model.
        - batch_size (int): Batch size for training.
        - input_shape (tuple[int, int]): Input shape of the data.
        - x_data (list[any]): Input data.
        - y_data (list[any]): Target data.
        - target_names (list[str], optional): Names of target classes.
        - class_mode (str, optional): Classification mode (default: 'categorical').
        - k_fold_n (int, optional): Number of folds in k-fold cross-validation (default: 5).
        - epochs (int, optional): Number of training epochs (default: 10).
        - callbacks (list[any], optional): List of callbacks for model training.
        - noise_levels (any, optional): Noise levels for data augmentation (default: NOISE_LEVELS).
        - save_models (bool, optional): Save trained models (default: False).
        - evaluate_models (bool, optional): Evaluate models during testing (default: False).
        - kfold_random_state (int, optional): K-Fold random state.
        """
        validate_kwargs(kwargs=kwargs, allowed_kwargs=NIGEP_AVAILABLE_KWARGS)
        self.execution_name: str = kwargs['execution_name']
        self.model: Sequential = kwargs['model']
        self.batch_size: int = kwargs['batch_size']
        self.input_shape: tuple[int, int] = kwargs['input_shape']
        self.x_data: list[any] = kwargs['x_data']
        self.y_data: list[any] = kwargs['y_data']
        self.target_names: list[str] = kwargs.get('target_names', None)
        self.class_mode: str = kwargs.get('class_mode', 'categorical')
        self.k_fold_n: int = kwargs.get('k_fold_n', 5)
        self.epochs: int = kwargs.get('epochs', 10)
        self.callbacks: list[any] = kwargs.get('callbacks', [])
        self.noise_levels: any = kwargs.get('noise_levels', NOISE_LEVELS)
        self.save_models: bool = kwargs.get('save_models', False)
        self.evaluate_models: bool = kwargs.get('evaluate_models', False)
        self.kfold_random_state: int = kwargs.get('kfold_random_state', 42)
        self.rw = ResultsWriter(self.execution_name)

    def __train_and_write_model(self, results_folder, fold_number, train_data, train_noise):
        print(f'Fold: {str(fold_number)} - Training with Noise: {str(train_noise)}')

        with self.lock:
            train_model(self.model, self.epochs, self.callbacks, train_data)
            write_model(results_folder, self.save_models, self.model, train_noise)

    def __test_and_write_metrics(self, results_folder, fold_number, test_index, train_noise):
        for test_noise in self.noise_levels:
            print(f'Fold: {str(fold_number)} - Train Noise: {str(train_noise)} - Test Noise: {str(test_noise)}')

            x_test, y_test = apply_indexed_noise(self.x_data, self.y_data, test_index, test_noise)

            if self.evaluate_models:
                self.model.evaluate(x_test, y_test)

            cm, cr = compute_metrics(self.model, self.class_mode, self.target_names, x_test, y_test)
            self.rw.write_new_metrics(results_folder, train_noise, test_noise, cr, cm, self.target_names)


    def __execute_fold(self, fold_number, train_index, test_index):
        results_folder = self.rw.write_k_subset_folder(fold_number)

        for train_noise in self.noise_levels:
            noised_data = apply_indexed_noise(self.x_data, self.y_data, train_index, train_noise)

            self.__train_and_write_model(results_folder, fold_number, noised_data, train_noise)

            self.__test_and_write_metrics(results_folder, fold_number, test_index, train_noise)

    def fit(self):
        kf = KFold(n_splits=self.k_fold_n, shuffle=True, random_state=self.kfold_random_state)
        dataset_splits = list(enumerate(kf.split(self.x_data, self.y_data)))

        with ThreadPoolExecutor(max_workers=2) as executor:
            futures = [executor.submit(
                self.__execute_fold, fold_number, train_index, test_index)
                for fold_number, (train_index, test_index)
                in dataset_splits
            ]

            for future in futures:
                future.result()

        self.rw.save_mean_merged_results()
        self.rw.save_heatmap_csv()

    def plot_and_save_generalization_profile(self):
        self.rw.plot_and_save_heatmap_png()



import numpy as np


def apply_indexed_noise(x_data, y_data, index, noise_amount):
    xx_data, yy_data = x_data[index], y_data[index]

    for idx, img in enumerate(xx_data):
        img_size = img.size
        noise_percentage = noise_amount
        noise_size = int(noise_percentage * img_size)
        random_indices = np.random.choice(img_size, noise_size)
        img_noised = img.copy()
        noise = np.random.choice([img.min(), img.max()], noise_size)
        img_noised.flat[random_indices] = noise
        xx_data[idx] = img_noised

    return xx_data, yy_data


def apply_noise(x_data, y_data, noise_amount):
    for idx, img in enumerate(x_data):
        img_size = img.size
        noise_percentage = noise_amount
        noise_size = int(noise_percentage * img_size)
        random_indices = np.random.choice(img_size, noise_size)
        img_noised = img.copy()
        noise = np.random.choice([img.min(), img.max()], noise_size)
        img_noised.flat[random_indices] = noise
        x_data[idx] = img_noised

    return x_data, y_data




from keras.layers import Layer
from keras import backend as K


class SaltAndPepperNoise(Layer):

    def __init__(self, ratio, **kwargs):
        super(SaltAndPepperNoise, self).__init__(**kwargs)
        self.supports_masking = True
        self.ratio = ratio

    @override
    def call(self, inputs):
        def noised():
            shp = K.shape(inputs)[1:]
            mask_select = K.random_bernoulli(shape=shp, p=self.ratio)
            mask_noise = K.random_bernoulli(shape=shp, p=0.5)
            out = inputs * (1 - mask_select) + mask_noise * mask_select
            return out

        return K.in_train_phase(noised(), inputs)

    def get_config(self):
        config = {'ratio': self.ratio}
        base_config = super(SaltAndPepperNoise, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



def get_results_columns(target_names):
    classes_precision_columns = []
    classes_recall_columns = []
    classes_f1score_columns = []
    for index, item in enumerate(target_names):
        classes_precision_columns.append(f'precision({item})')
        classes_recall_columns.append(f'recall({item})')
        classes_f1score_columns.append(f'f1-score({item})')

    return [
        'train-dataset-noise', 'test-dataset-noise',
        *classes_precision_columns,
        'precision(macro-avg)', 'precision(weighted-avg)',
        *classes_recall_columns,
        'recall(macro-avg)',  'recall(weighted-avg)',
        *classes_f1score_columns,
        'f1-score(accuracy)', 'f1-score(macro-avg)', 'f1-score(weighted-avg)'
    ]


def validate_kwargs(
    kwargs, allowed_kwargs, error_message="Keyword argument not understood:"
):
    """Checks that all keyword arguments are in the set of allowed keys."""
    for kwarg in kwargs:
        if kwarg not in allowed_kwargs:
            raise TypeError(error_message, kwarg)


def write_model(results_folder, save_model, model, noise):
    if save_model:
        model.save(f'{results_folder}/train_{noise}.keras')




from sklearn.base import BaseEstimator, TransformerMixin

from ..lib.apply_noise import apply_indexed_noise
from ..lib.metrics import compute_metrics
from ..lib.train_model import train_model
from ..lib.functions import write_model


class NigepModel(BaseEstimator, TransformerMixin):
    def __init__(self,
                 results_folder,
                 fold_number,
                 train_noise,
                 train_index,
                 model,
                 epochs,
                 callbacks,
                 class_mode,
                 target_names,
                 save_models,
                 evaluate_models=True):
        self.results_folder = results_folder
        self.train_index = train_index
        self.train_noise = train_noise
        self.model = model
        self.fold_number = fold_number
        self.epochs = epochs
        self.callbacks = callbacks
        self.class_mode = class_mode
        self.target_names = target_names
        self.evaluate_models = evaluate_models
        self.save_models = save_models

    def fit(self, X, y=None):
        noised_data = apply_indexed_noise(X, y, self.train_index, self.train_noise)

        print(f'Fold: {str(self.fold_number)} - Training with Noise: {str(self.train_noise)}')

        train_model(self.model, self.epochs, self.callbacks, noised_data)
        write_model(self.results_folder, self.save_models, self.model, self.train_noise)

        return self

    def evaluate_model(self, X, y, test_index, test_noise):
        print(f'Fold: {str(self.fold_number)} - Train Noise: {str(self.train_noise)} - Test Noise: {str(test_noise)}')

        x_test, y_test = apply_indexed_noise(X, y, test_index, test_noise)

        if self.evaluate_models:
            self.model.evaluate(x_test, y_test)

        cm, cr = compute_metrics(self.model, self.class_mode, self.target_names, x_test, y_test)

        return cm, cr

